from .FunctionAnnotationChecker import (
    InvalidArgumentTypeException,
    InvalidReturnTypeException,
    annotation_checker,
)


__all__ = ["InvalidArgumentTypeException",
          "InvalidReturnTypeException",
          "annotation_checker"]




import time
import numpy as np

_print_orig = print


def timer(iteration: int = 1, result_index: int = 0, print_only_first: bool = True):
    def _timer(func):
        def wrapper(*args, **kwargs):
            result = []
            result_append = result.append
            dt = []
            dt_append = dt.append

            def do_nothing(*args, **kwargs):
                pass

            for i in range(iteration):
                if print_only_first and i > 0:
                    globals()["print"] = do_nothing

                st = time.time()
                result_append(func(*args, **kwargs))
                et = time.time()
                dt_append(et - st)

            globals()["print"] = _print_orig

            dt = np.array(dt)

            print(f"Average process time: {np.average(dt)}")
            print(f"Max process time    : {dt.max()}")
            print(f"Minumin process time: {dt.min()}")

            return result[result_index]

        return wrapper

    return _timer



from typing import Union, Any
import inspect


class InvalidArgumentTypeException(Exception):
    def __init__(self, arg_name: str, valid_type: Any, value: Any):
        mes = f"{arg_name} requires type {valid_type}. You specified the argument as {value}"
        super().__init__(mes)


class InvalidReturnTypeException(Exception):
    def __init__(self, valid_type: Any, value: Any):
        mes = f"Return values requires type {valid_type}. Function tried ti return values as {value}"
        super().__init__(mes)


def _get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def annotation_checker(func):
    annotations = func.__annotations__
    return_type = annotations["return"]
    func_varnames = list(func.__code__.co_varnames[: func.__code__.co_argcount])
    func_varnames_keys = {k: i for i, k in enumerate(func_varnames) if k in annotations}

    def wrapper(*args, **kwargs):
        annotations_list = [k for k in annotations.keys()]
        annotations_list.remove("return")

        for k, v in kwargs.items():
            if k in annotations:
                if not isinstance(v, annotations[k]):
                    raise InvalidArgumentTypeException(k, annotations[k], type(v))

                annotations_list.remove(k)

        default_args = _get_default_args(func)
        for annotation in annotations_list:
            if (
                func_varnames_keys[annotation] >= len(args)
                and annotation in default_args
            ):
                if not isinstance(default_args[annotation], annotations[k]):
                    raise InvalidArgumentTypeException(
                        annotation, annotations[annotation], default_args[annotation]
                    )
            else:
                value = args[func_varnames_keys[annotation]]
                if not isinstance(value, annotations[annotation]):
                    raise InvalidArgumentTypeException(
                        annotation, annotations[annotation], type(value)
                    )

        result = func(*args, **kwargs)
        result_type = type(result)

        # In the case of return value is Union
        if "Union" in str(return_type):
            args = return_type.__args__
            if not result_type in args:
                raise InvalidReturnTypeException(args, return_type)
            assert result_type in args

        else:
            assert result_type == return_type

        return result

    return wrapper



from ..FunctionAnnotationChecker.FunctionAnnotationChecker import _get_default_args
class InvalidExtensionException(Exception):
    """InvalidExtensionException will be raised when arguments file extension is not equal to specified extension."""

    def __init__(self, arg_name: str, valid_extension: str, value: str):
        """
        Arguments:
        ----------
            arg_name {str} -- argument name
            valid_extension {str} -- extension which you want to fixed
            value {str} -- argument value
        """
        msg = f"{arg_name} requires extension {valid_extension}. You specified the argument as {value}"
        super().__init__(msg)


def _sort_argument(*args, **kwargs):
    argument_names = kwargs.get("argument_names")
    del kwargs["argument_names"]

    result = {}
    for arg, name in zip(args, argument_names):
        if name == "self":
            continue

        result[name] = arg

    for k, v in kwargs.items():
        if not k in result:
            result[k] = v

    return result


def extension_checker(**vkwargs):
    """extension_checker checks arguments' file name extension.

    Examples:
    ---------
        >>> @extension_checker(filename="jpg)
        ... def hoge(filename):
        ...     print(filename)
        >>> hoge("hoge.jpg")
        hoge.jpg
        >>> hoge(filename="hoge.jpg")
        hoge.jpg
    """

    accept_none = vkwargs.get("accept_none", False)
    def _check_filename(func):
        func_varnames = list(func.__code__.co_varnames[: func.__code__.co_argcount])
        args_index = {k: i for i, k in enumerate(func_varnames)}
        default_args = _get_default_args(func)

        def wrapper(*args, **kwargs):
            kwargs.update(default_args)
            arguments = _sort_argument(*args, **kwargs, argument_names=list(func.__code__.co_varnames))

            for varname, extension in vkwargs.items():
                if varname == "accept_none":
                    continue

                if arguments[varname] is None and accept_none:
                    continue

                ext = arguments[varname].split(".")[-1]
                if not ext == extension:
                    raise InvalidExtensionException(
                        varname, extension, arguments[varname]
                    )


            if func_varnames[0] == "self":
                arguments["self"] = args[0]

            return func(**arguments)
            # return func(*args, **kwargs)

        return wrapper

    return _check_filename



from glob import glob
from os.path import basename, splitext
from setuptools import setup, find_packages

def _get_requires_from_file(filename):
    return open(filename).read().splitlines()


setup(
    name="udeco",
    version="1.0.2",
    license="MIT",
    author="Akagawa Daisuke",
    url="http://github.com/Akasan/udeco",
    packages=["udeco",
              "udeco.FileExtensionChecker",
              "udeco.TimeRecorder",
              "udeco.FunctionAnnotationChecker"],
    include_package_data=True,
    zip_safe=False,
)
